@@ -18,16 +18,27 @@ from posixpath import normpath
 import six
 from six.moves import urllib
 from six.moves.urllib.parse import quote, unquote
-try:
-    from icecream import ic
-except ImportError:  # Graceful fallback if IceCream isn't installed.
-    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)
 
 from .omdict1D import omdict1D
 from .compat import string_types, UnicodeMixin
 from .common import (
     callable_attr, is_iterable_but_not_string, absent as _absent)
 
+#
+# TODO(grun): Subclass Path, PathCompositionInterface, Query, and
+# QueryCompositionInterface into two subclasses each - one for the URL
+# and one for the Fragment.
+#
+# Subclasses will clean up the code because the valid encodings are
+# different between a URL Path and a Fragment Path and a URL Query and a
+# Fragment Query.
+#
+# For example, '?' and '#' don't need to be encoded in Fragment Path
+# segments but must be encoded in URL Path segments.
+#
+# Similarly, '#' doesn't need to be encoded in Fragment Query keys and
+# values, but must be encoded in URL Query keys and values.
+#
 
 # Map of various URL schemes to their default ports. Scheme strings are
 # lowercase.
@@ -206,12 +217,6 @@ def set_scheme(url, scheme):
         return '%s:%s' % (scheme, after_scheme)
 
 
-# 'netloc' in Python parlance, 'authority' in RFC 3986 parlance.
-def has_netloc(url):
-    scheme = get_scheme(url)
-    return url.startswith('//' if scheme is None else scheme + '://')
-
-
 def urlsplit(url):
     """
     Parameters:
@@ -251,7 +256,6 @@ def urlsplit(url):
         netloc = None
 
     scheme = original_scheme
-
     return urllib.parse.SplitResult(scheme, netloc, path, query, fragment)
 
 
@@ -263,30 +267,16 @@ def urljoin(base, url):
 
     Returns: The resultant URL from joining <base> and <url>.
     """
-    base_scheme = get_scheme(base) if has_netloc(base) else None
-    url_scheme = get_scheme(url) if has_netloc(url) else None
+    if not base:
+        return url
 
-    if base_scheme is not None:
-        # For consistent URL joining, switch the base URL's scheme to
-        # 'http'. urllib.parse.urljoin() behaves differently depending on the
-        # scheme. E.g.
-        #
-        #   >>> urllib.parse.urljoin('http://google.com/', 'hi')
-        #   'http://google.com/hi'
-        #
-        # vs
-        #
-        #   >>> urllib.parse.urljoin('asdf://google.com/', 'hi')
-        #   'hi'
-        root = set_scheme(base, 'http')
-    else:
-        root = base
+    base_scheme = get_scheme(base)
+    url_scheme = get_scheme(url)
 
-    joined = urllib.parse.urljoin(root, url)
-
-    new_scheme = url_scheme if url_scheme is not None else base_scheme
-    if new_scheme is not None and has_netloc(joined):
-        joined = set_scheme(joined, new_scheme)
+    http_base = set_scheme(base, 'http')
+    joined = urllib.parse.urljoin(http_base, url)
+    if not url_scheme:
+        joined = set_scheme(joined, base_scheme)
 
     return joined
 
@@ -310,7 +300,6 @@ def join_path_segments(*args):
     Returns: A list containing the joined path segments.
     """
     finals = []
-
     for segments in args:
         if not segments or segments == ['']:
             continue
@@ -325,7 +314,6 @@ def join_path_segments(*args):
             elif finals[-1] != '' and segments[0] == '' and len(segments) > 1:
                 segments = segments[1:]
             finals.extend(segments)
-
     return finals
 
 
@@ -471,7 +459,6 @@ class Path(object):
             segments.insert(0, '')
 
         self.load(join_path_segments(segments, newsegments))
-
         return self
 
     def set(self, path):
@@ -488,7 +475,6 @@ class Path(object):
                 segments = self._segments_from_path(path)
             base = ([''] if self.isabsolute else []) + self.segments
             self.load(remove_path_segments(base, segments))
-
         return self
 
     def normalize(self):
@@ -502,7 +488,6 @@ class Path(object):
             if normalized.startswith('//'):  # http://bugs.python.org/636648
                 normalized = '/' + normalized.lstrip('/')
             self.load(normalized)
-
         return self
 
     def asdict(self):
@@ -870,7 +855,6 @@ class Query(object):
             else:
                 key = item
                 self.params.pop(key, None)
-
         return self
 
     @property
@@ -1131,7 +1115,6 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.path.add(path)
         if args is not _absent:
             self.query.add(args)
-
         return self
 
     def set(self, path=_absent, args=_absent, separator=_absent):
@@ -1141,7 +1124,6 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.query.load(args)
         if separator is True or separator is False:
             self.separator = separator
-
         return self
 
     def remove(self, fragment=_absent, path=_absent, args=_absent):
@@ -1151,7 +1133,6 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
             self.path.remove(path)
         if args is not _absent:
             self.query.remove(args)
-
         return self
 
     def asdict(self):
@@ -1187,9 +1168,9 @@ class Fragment(FragmentPathCompositionInterface, QueryCompositionInterface):
         if path and (not query or not self.separator):
             path = path.replace('%3F', '?')
 
-        separator = '?' if path and query and self.separator else ''
-
-        return path + separator + query
+        if query and path:
+            return path + ('?' if self.separator else '') + query
+        return path + query
 
     def __repr__(self):
         return "%s('%s')" % (self.__class__.__name__, str(self))
@@ -1298,7 +1279,6 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         self.path.load(tokens.path)
         self.query.load(tokens.query)
         self.fragment.load(tokens.fragment)
-
         return self
 
     @property
@@ -1430,7 +1410,6 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         if self.port and self.port != DEFAULT_PORTS.get(self.scheme):
             port = ':%s' % self.port
         origin = '%s://%s%s' % (scheme, host, port)
-
         return origin
 
     @origin.setter
@@ -1500,7 +1479,6 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             self.query.add(query_params)
         if fragment_path is not _absent or fragment_args is not _absent:
             self.fragment.add(path=fragment_path, args=fragment_args)
-
         return self
 
     def set(self, args=_absent, path=_absent, fragment=_absent, scheme=_absent,
